""" Events that the Command Centre

Command Centre has about 80 event types that occur when somebody 
authenticates at a device, usually by badging a card.
"""

from ..core import (
    Capabilities,
    APIEndpoint,
    EndpointConfig,
)

from ...dto.detail import (
    EventDetail,
)

from ...dto.response import (
    EventTypeResponse,
    EventSummaryResponse,
)


class Event(APIEndpoint):
    """Event
    
    How updates differ from GET:

    - events.updates.href as found in /api discovery, it will give you the 
      first events that arrive after you call the endpoint.
    - the url in updates in /api/events or /api/events/updates, returns
      events that arrive at the bookmark and meet the search criteria.

    the next href is a non blocking call to get the next set of events.
    """

    @classmethod
    async def get_config(cls) -> EndpointConfig:
        return EndpointConfig(
            endpoint=Capabilities.CURRENT.features.events.events,
            endpoint_follow=Capabilities.CURRENT.features.events.updates,
            dto_follow=EventSummaryResponse,
            dto_list=EventSummaryResponse,
            dto_retrieve=EventDetail,
        )

class EventType(APIEndpoint):
    """EventType"""

    @classmethod
    async def get_config(cls) -> EndpointConfig:
        return EndpointConfig(
            endpoint=Capabilities.CURRENT.features.events.event_groups,
            dto_list=EventTypeResponse,
        )


class EventGroups(APIEndpoint):
    """ Event Groups are used to filter events
    
    Use this to dynamically discover a list of event groups which you can
    then use for various updates from the command centre.

    See above Event class for more details.
    """

    @classmethod
    async def get_config(cls) -> EndpointConfig:
        return EndpointConfig(
            endpoint=Capabilities.CURRENT.features.events.event_groups,
            dto_list=EventTypeResponse,
        )



""" Command Centre API discovery

The Command Centre API follows the HATEOAS principle, and all clients are meant
to discover the API by following the links provided by the server.

An instance of this class is used across the library to provide to dynamically
link to the API endpoints.

WARNING: please do not hardcode any URLs in the code, if you see this then please
report this as a bug.
"""

from ..utils import (
    AppBaseModel,
    OptionalHrefMixin,
)

class FeatureAccessGroupsRef(AppBaseModel):
    access_groups: OptionalHrefMixin = None

class FeatureAccessZonesRef(AppBaseModel):
    access_zones: OptionalHrefMixin = None

class FeatureAlarmZonesRef(AppBaseModel):
    alarm_zones: OptionalHrefMixin = None

class FeatureAlarmsRef(AppBaseModel):
    alarms: OptionalHrefMixin = None
    divisions: OptionalHrefMixin = None
    updates: OptionalHrefMixin = None

class FeatureCardTypesRef(AppBaseModel):
    assign: OptionalHrefMixin = None
    card_types: OptionalHrefMixin = None

class FeatureCardholdersRef(AppBaseModel):
    cardholders: OptionalHrefMixin = None
    changes: OptionalHrefMixin = None
    update_location_access_zones: OptionalHrefMixin = None

class FeatureCompetenciesRef(AppBaseModel):
    competencies: OptionalHrefMixin = None

class FeatureDayCategoriesRef(AppBaseModel):
    day_categories: OptionalHrefMixin = None

class FeatureDivisionsRef(AppBaseModel):
    divisions: OptionalHrefMixin = None

class FeatureDoorsRef(AppBaseModel):
    doors: OptionalHrefMixin = None

class FeatureElevatorsRef(AppBaseModel):
    elevator_groups: OptionalHrefMixin = None

class FeatureEventsRef(AppBaseModel):
    divisions: OptionalHrefMixin = None
    event_groups: OptionalHrefMixin = None
    events: OptionalHrefMixin = None
    updates: OptionalHrefMixin = None

class FeatureFenceZonesRef(AppBaseModel):
    fence_zones: OptionalHrefMixin = None

class FeatureInputsRef(AppBaseModel):
    inputs: OptionalHrefMixin = None

class FeatureInterlockGroupsRef(AppBaseModel):
    interlock_groups: OptionalHrefMixin = None

class FeatureItemsRef(AppBaseModel):
    item_types: OptionalHrefMixin = None
    items: OptionalHrefMixin = None
    updates: OptionalHrefMixin = None

class FeatureLockerBanksRef(AppBaseModel):
    locker_banks: OptionalHrefMixin = None

class FeatureMacrosRef(AppBaseModel):
    macros: OptionalHrefMixin = None

class FeatureOperatorGroupsRef(AppBaseModel):
    operator_groups: OptionalHrefMixin = None

class FeatureOutputsRef(AppBaseModel):
    outputs: OptionalHrefMixin = None

class FeaturePersonalDataFieldsRef(AppBaseModel):
    personal_data_fields: OptionalHrefMixin = None

class FeatureReceptionsRef(AppBaseModel):
    receptions: OptionalHrefMixin = None

class FeatureRolesRef(AppBaseModel):
    roles: OptionalHrefMixin = None

class FeatureSchedulesRef(AppBaseModel):
    schedules: OptionalHrefMixin = None

class FeatureVisitsRef(AppBaseModel):
    visits: OptionalHrefMixin = None



""" Decorators for annotating functions for behaviours.

    All decorators are defined here and are referenced across the codebase.
    Note that some of them reference the bootstrapped environment where
    information is discovered from the live Command Centre the instance
    is connected to.

    See Also:
        https://github.com/anomaly/gallagher/issues/69
"""

def deprecated(func):
    """Decorator to indicate that a function is deprecated.

    This decorator is used to indicate that a function is deprecated and
    should not be used. It will raise a warning when the function is called.

    Args:
        func (Callable): The function to be deprecated.

    Returns:
        Callable: The wrapped function.

    Examples:
        >>> @deprecated
        ... def old_function():
        ...     pass

    """
    import warnings
    import functools

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        warnings.warn(
            f"Call to deprecated function {func.__name__}.", 
            category=DeprecationWarning
        )
        return func(*args, **kwargs)

    return wrapper


def available(version_string):
    """Decorator to indicate that a function is available in a specific version.

    This decorator is used to indicate that a function is available in a specific
    version of the Command Centre API. It will raise an error if the function is
    called in a version that does not support it.

    Args:
        version_string (str): The version string the function is available in.

    Returns:
        Callable: The wrapped function.

    Examples:
        >>> @available("1.0.0")
        ... def new_function():
        ...     pass

    """
    import functools

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            from .cc.core import Capabilities

            if Capabilities.CURRENT.version < version_string:
                raise RuntimeError(
                    f"Function {func.__name__} is not available in version {Capabilities.CURRENT.version}."
                )

            return func(*args, **kwargs)

        return wrapper

    return decorator


from typing import Optional

from ..utils import (
    AppBaseModel,
)

from ..ref import (
    FeatureAccessGroupsRef,
    FeatureAccessZonesRef,
    FeatureAlarmZonesRef,
    FeatureAlarmsRef,
    FeatureCardTypesRef,
    FeatureCardholdersRef,
    FeatureCompetenciesRef,
    FeatureDayCategoriesRef,
    FeatureDivisionsRef,
    FeatureDoorsRef,
    FeatureElevatorsRef,
    FeatureEventsRef,
    FeatureFenceZonesRef,
    FeatureInputsRef,
    FeatureInterlockGroupsRef,
    FeatureItemsRef,
    FeatureLockerBanksRef,
    FeatureMacrosRef,
    FeatureOperatorGroupsRef,
    FeatureOutputsRef,
    FeaturePersonalDataFieldsRef,
    FeatureReceptionsRef,
    FeatureRolesRef,
    FeatureSchedulesRef,
    FeatureVisitsRef,
)


class SemVerDetail(AppBaseModel):
    """Semantic Versioning details
    
    As defined by https://semver.org/ and implemented by the API
    the API client sends this back for the API and the command centre
    """
    major: int
    minor: int
    revision: int
    build: int


class VersionsDetail(AppBaseModel):
    """A detailed list of versions that are available on the server"""
    command_centre: Optional[SemVerDetail] = None
    api: Optional[SemVerDetail] = None

class FeaturesDetail(
    AppBaseModel,
):
    """A detailed list of features that are available on the server.

    All features are marked as Optional, which means that by default
    it's assumed that they are not available on the server. Upon discovery
    if a feature is enabled on the server then we receive a href which
    indicates to the client that the feature is available.

    If a feature is unavailable the API client will throw an exception.
    """

    access_groups: Optional[FeatureAccessGroupsRef] = FeatureAccessGroupsRef()
    access_zones: Optional[FeatureAccessZonesRef] = FeatureAccessZonesRef()
    alarm_zones: Optional[FeatureAlarmZonesRef] = FeatureAlarmZonesRef()
    alarms: Optional[FeatureAlarmsRef] = FeatureAlarmsRef()
    card_types: Optional[FeatureCardTypesRef] = FeatureCardTypesRef()
    cardholders: Optional[FeatureCardholdersRef] = FeatureCardholdersRef()
    competencies: Optional[FeatureCompetenciesRef] = FeatureCompetenciesRef()
    day_categories: Optional[FeatureDayCategoriesRef] = FeatureDayCategoriesRef()
    divisions: Optional[FeatureDivisionsRef] = FeatureDivisionsRef()
    doors: Optional[FeatureDoorsRef] = FeatureDoorsRef()
    elevators: Optional[FeatureElevatorsRef] = FeatureElevatorsRef()
    events: Optional[FeatureEventsRef] = FeatureEventsRef()
    fence_zones: Optional[FeatureFenceZonesRef] = FeatureFenceZonesRef()
    inputs: Optional[FeatureInputsRef] = FeatureInputsRef()
    interlock_groups: Optional[FeatureInterlockGroupsRef] = FeatureInterlockGroupsRef()
    items: Optional[FeatureItemsRef] = FeatureItemsRef()
    locker_banks: Optional[FeatureLockerBanksRef] = FeatureLockerBanksRef()
    macros: Optional[FeatureMacrosRef] = FeatureMacrosRef()
    operator_groups: Optional[FeatureOperatorGroupsRef] = FeatureOperatorGroupsRef()
    outputs: Optional[FeatureOutputsRef] = FeatureOutputsRef()
    personal_data_fields: Optional[FeaturePersonalDataFieldsRef] = (
        FeaturePersonalDataFieldsRef()
    )
    receptions: Optional[FeatureReceptionsRef] = FeatureReceptionsRef()
    roles: Optional[FeatureRolesRef] = FeatureRolesRef()
    schedules: Optional[FeatureSchedulesRef] = FeatureSchedulesRef()
    visits: Optional[FeatureVisitsRef] = FeatureVisitsRef()



from ..utils import (
    AppBaseResponseModel,
    AppBaseResponseWithFollowModel,
)

from ..summary import (
    EventGroupSummary,
    EventSummary,
)

class EventSummaryResponse(
    AppBaseResponseWithFollowModel,
):
    """ Event Summary Response
    """

    events: list[EventSummary]

class EventTypeResponse(
    AppBaseResponseModel,
):
    """Event Type Response

    Event Type Responses return a set of eventGroups which in turn
    has identifiers, names and event types.
    """

    event_groups: list[EventGroupSummary]

